/*	The purpose of this program is to demonstrate,
 * that 32-bit integer Math can be used to approximate
 * a sine function on CPUs with no floating-point unit.
 *
 * To do so, a 16-bit parameter will be read in, the
 * 8 MSBs of which will be used to look up a sine value
 * from a table.
 *
 * However, due to the fact that the first derivative
 * is the cosine, and that the second derivative is the negative
 * of the sine, a closer approximation will make a 2nd and
 * a 3rd lookup in the table.
 *
 * 	This is the third version of the exercise, which has been
 * conceived more, to emulate the requirements of a weak
 * Digital Signal Processor, the bit-alignment of which
 * might be friendly with 16-bit words...
 *
 * The table will only contain 128 / 256 elements, the second
 * set of 128 of which will be derived as the negatives of
 * the first 128. This will allow the table itself only to
 * take up 256 bytes of RAM.
 *
 */

#include <stdio.h>
#include <stdint.h>

static const uint16_t table[128] =

#include "table3.h"

int sin_lookup(int index)
{
    int c = 0;

    c = table[index & 0x7f];
    return (index & 0x80) ? - c : c;
}

int cos_lookup(int index)
{
    int c = 0;

    c = (index + 0x40) & 0xff;
    return sin_lookup(c);
}

int sine(int param)
{
    const int pi_2 = 411774;	//  A 19-bit value.
    int index = 0;
    int frac = 0;
    int temp = 0;

    index = (param & 0xff00) >> 8;
    frac = param & 0xff;

    //  Note: The fractional part's significance is
    //  that, of a 16-bit fraction. After being
    //  multiplied by pi_2, and then right-shifted,
    //  its max value will have 11 bits...

    frac *= pi_2;	//  Max value has 27 bits.
    frac += 0x8000;
    frac = frac >> 16;
    //  Down to 11 bits.

    temp = sin_lookup(index);

    temp += ((cos_lookup(index) * frac) + 0x8000) >> 16;

    frac *= frac;	//  Max value has 22 bits.
    frac += 0x8000;
    frac = frac >> 16;
    //  Down to 6 bits.

    temp -= ((sin_lookup(index) * frac) + 0x10000) >> 17;
    //  Inentionally halving.

    return temp;
}

int cosine(int param)
{
    return sine((param + 0x4000) & 0xffff);
}

/*	I guess that the real test will be, to cheat and use the standard
 * Math library, to compute what the values should be,
 * and then, to compute the variance between the integer
 * implementation and the established values...
 */

#include <math.h>

int main(int argc, char* argv[])
{
    double sum__x_2 = 0.0;
    double sum_x = 0.0;
    double variance = 0.0;

    int estab = 0;
    double exact = 0.0;

    int test = 0;

    for (int i = 0; i < 65536; i++) {
        exact = sin(2.0 * M_PI * i / 65536.0);
        estab = floor((exact * 32768.0) + 0.5);

        test = sine(i);

        sum__x_2 += (test - estab) * (test - estab);
        sum_x += (test - estab);
    }

    sum__x_2 /= 65536.0;
    sum_x /= 65536.0;
    variance = sum__x_2 - (sum_x * sum_x);


    printf("The variance within the differences between the\n");
    printf("integer-Math and the established results was:\n");
    printf("%.6f\n\n", variance);

    printf("This value is in integer units, squared, then averaged,\n");
    printf("that were computed as signed 16-bit fractions,\n");
    printf("for all possible 16-bit parameters, defining 1 cycle.\n");
    printf("The maximum amplitudes were to Include, +/- 32768\n\n");

    printf("The DC offset was: %.6f\n\n", sum_x);

    return 0;
}
